/*
 * fsarchiver: Filesystem Archiver
 *
 * Copyright (C) 2008-2009 Francois Dupoux.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License v2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * Homepage: http://www.fsarchiver.org
 */

#include "fsarchiver.h"

#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <stdlib.h>
#include <dirent.h>
#include <sys/stat.h>
#include <endian.h>
#include <unistd.h>
#include <byteswap.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/mount.h>
#include <sys/statvfs.h>
#include <attr/xattr.h>
#include <linux/limits.h>
#include <zlib.h>

#include "dico.h"
#include "dichl.h"
#include "archive.h"
#include "common.h"
#include "create.h"
#include "md5.h"
#include "strlist.h"
#include "filesys.h"
#include "fs_ext2.h"
#include "fs_xfs.h"
#include "fs_reiserfs.h"
#include "fs_reiser4.h"
#include "fs_jfs.h"
#include "fs_btrfs.h"
#include "fs_ntfs.h"
#include "thread_comp.h"
#include "thread_archio.h"
#include "syncthread.h"
#include "regmulti.h"
#include "crypto.h"

struct s_filesysdat
{
	struct s_dichl		*dichardlinks;
	struct s_regmulti	regmulti;
	int			fstype;
	int			fsid;
};

int backup_obj_regfile_multi(struct s_archive *ai, struct s_dico *header, char *relpath, char *fullpath, u64 filesize, u16 fsid, struct s_regmulti *m)
{
	char databuf[FSA_MAX_SMALLFILESIZE];
	struct md5_ctx md5ctx;
	u8 md5sum[16];
	int ret=0;
	int res;
	int fd;
	
	// The checksum will be in the obj-header not in a file footer
	
	if ((fd=open64(fullpath, O_RDONLY|O_LARGEFILE))<0)
	{	sysprintf("Cannot open small file %s for reading\n", relpath);
		return -1;
	}
	
	msgprintf(MSG_DEBUG1, "backup_obj_regfile_multi(file=%s, size=%lld)\n", relpath, (long long)filesize);
	
	res=read(fd, databuf, (long)filesize);
	close(fd);
	if (res!=filesize)
	{	sysprintf("Cannot read data block size=%ld from small file %s, res=%ld\n", (long)filesize, relpath, (long)res);
		return -1;
	}
	
	md5_init_ctx(&md5ctx);
	md5_process_bytes(databuf, filesize, &md5ctx);
	md5_finish_ctx(&md5ctx, md5sum);
	
	dico_add_data(header, 0, DISKITEMKEY_MD5SUM, md5sum, 16);
	
	// if shared-block with many small files is full, push it to queue and make a new one
	if (regmulti_save_enough_space_for_new_file(m, filesize)==false)
	{
		if (regmulti_save_enqueue(m, &g_queue, fsid)!=0)
		{	errprintf("Cannot queue last block of small-files\n");
			return -1;
		}
		
		regmulti_empty(m);
	}
	
	// copy current small file to the shared-block
	if (regmulti_save_addfile(m, header, databuf, filesize)!=0)
	{	errprintf("Cannot add small-file %s to regmulti structure\n", relpath);
		return -1;
	}
	
	return ret;
}

int backup_obj_regfile_unique(struct s_archive *ai, struct s_dico *header, char *relpath, char *fullpath, u64 filesize, u16 fsid)
{
	struct s_dico *footerdico=NULL;
	struct s_blockinfo blkinfo;
	struct md5_ctx md5ctx;
	u32 curblocksize;
	u64 remaining;
	char text[256];
	u8 *origblock;
	u8 md5sum[16];
	u64 filepos;
	int ret=0;
	int res;
	int fd;
	
	// init
	if ((footerdico=dico_alloc())==NULL)
	{	errprintf("dico_alloc() failed\n");
		return -1;
	}
	
	// write header with file attributes
	queue_add_header(&g_queue, header, FSA_MAGIC_OBJT, fsid);
	
	md5_init_ctx(&md5ctx);
	
	if ((fd=open64(fullpath, O_RDONLY|O_LARGEFILE))<0)
	{	sysprintf("Cannot open %s for reading\n", relpath);
		return -1;
	}
	
	msgprintf(MSG_DEBUG1, "backup_obj_regfile_unique(file=%s, size=%lld)\n", relpath, (long long)filesize);
	for (filepos=0; (filepos < filesize) && (get_interrupted()==false); filepos+=curblocksize)
	{
		remaining=filesize-filepos;
		curblocksize=min(remaining, g_options.datablocksize);
		msgprintf(MSG_DEBUG2, "----> filepos=%lld, remaining=%lld, curblocksize=%lld\n", (long long)filepos, (long long)remaining, (long long)curblocksize);
		
		origblock=malloc(curblocksize);
		if (!origblock)
		{	errprintf("malloc(%ld) failed: cannot allocate data block\n", (long)origblock);
			ret=-1;
			goto backup_obj_regfile_unique_error;
		}
		
		if ((res=read(fd, origblock, (long)curblocksize))!=curblocksize)
		{	sysprintf("Cannot read data block from %s, block=%ld and res=%ld\n", relpath, (long)curblocksize, (long)res);
			ret=-1;
			goto backup_obj_regfile_unique_error;
		}
		
		md5_process_bytes(origblock, curblocksize, &md5ctx);
		
		// add block to the queue
		memset(&blkinfo, 0, sizeof(blkinfo));
		blkinfo.blkrealsize=curblocksize;
		blkinfo.blkdata=(char*)origblock;
		blkinfo.blkoffset=filepos;
		blkinfo.blkfsid=fsid;
		if (queue_add_block(&g_queue, &blkinfo, QITEM_STATUS_TODO)!=0)
		{	sysprintf("queue_add_block(%s) failed\n", relpath);
			ret=-1;
			goto backup_obj_regfile_unique_error;
		}
	}
	
	if (get_interrupted()==true)
	{	errprintf("operation has been interrupted\n");
		ret=-1;
		goto backup_obj_regfile_unique_error;
	}
	
	// write the footer with the global md5sum
	md5_finish_ctx(&md5ctx, md5sum);
	msgprintf(MSG_DEBUG1, "--> finished loop for file=%s, size=%lld, md5=[%s]\n", relpath, (long long)filesize, format_md5(text, sizeof(text), md5sum));
	dico_add_data(footerdico, 0, BLOCKFOOTITEMKEY_MD5SUM, md5sum, 16);
	if (queue_add_header(&g_queue, footerdico, FSA_MAGIC_FILF, fsid)!=0)
	{	msgprintf(2, "Cannot write footer for file %s\n", relpath);
		ret=-1;
		goto backup_obj_regfile_unique_error;
	}
	
backup_obj_regfile_unique_error:
	close(fd);
	return ret;
}

int backup_item_xattr(struct s_archive *ai, char *root, char *relpath, struct stat64 *statbuf, struct s_dico *d)
{
	char fullpath[PATH_MAX];
	char buffer[4096];
	char value[1024];
	int valsize=0;
	int listlen;
	u64 attrcnt;
	int pos;
	int len;
	
	// init
	concatenate_paths(fullpath, sizeof(fullpath), root, relpath);
	attrcnt=0;
	
	memset(buffer, 0, sizeof(buffer));
	listlen=llistxattr(fullpath, buffer, sizeof(buffer)-1);
	msgprintf(MSG_DEBUG2, "llistxattr(%s)=%d\n", relpath, listlen);
	
	for (pos=0; (pos<listlen) && (pos<sizeof(buffer)); pos+=len)
	{
		len=strlen(buffer+pos)+1;
		memset(value, 0, sizeof(value));
		msgprintf(MSG_VERB2, "            lgetxattr(%s,%s)=%d: [%s]=[%.8x]\n", relpath, buffer+pos, valsize, buffer+pos, *((u64*)&value));
		valsize=lgetxattr(fullpath, buffer+pos, value, sizeof(value));
		if (valsize>=0)
		{
			msgprintf(MSG_DEBUG2, "            lgetxattr(%s)=%d: [%s]=[%s]\n", relpath, valsize, buffer+pos, value);
			msgprintf(MSG_DEBUG2, "            dico_add_string(%s, xattr): key=%d, name=[%s]\n", relpath, (int)(2*attrcnt)+0, buffer+pos);
			dico_add_string(d, DICO_OBJ_SECTION_XATTR, (2*attrcnt)+0, buffer+pos);
			msgprintf(MSG_DEBUG2, "            dico_add_string(%s, xattr): key=%d, valsize=[%d]\n", relpath, (int)(2*attrcnt)+1, valsize);
			dico_add_data(d, DICO_OBJ_SECTION_XATTR, (2*attrcnt)+1, value, valsize);
			attrcnt++;
		}
	}
	
	return 0;
}

int backup_item_winattr(struct s_archive *ai, char *root, char *relpath, struct stat64 *statbuf, struct s_dico *d)
{
	char fullpath[PATH_MAX];
	char value[1024];
	int valsize=0;
	u64 attrcnt;
	int i;
	
	char *winattr[]= {"system.ntfs_acl", "system.ntfs_attrib", "system.ntfs_reparse_data", NULL};
	
	// init
	concatenate_paths(fullpath, sizeof(fullpath), root, relpath);
	attrcnt=0;
	
	for (i=0; winattr[i]; i++)
	{
		memset(value, 0, sizeof(value));
		msgprintf(MSG_VERB2, "            lgetxattr(%s,%s)=%d: [%s]\n", relpath, winattr[i], valsize, winattr[i]);
		valsize=lgetxattr(fullpath, winattr[i], value, sizeof(value));
		if (valsize>=0)
		{
			msgprintf(MSG_VERB2, "            dico_add_string(%s, winattr): key=%d, name=[%s]\n", relpath, (int)(2*attrcnt)+0, winattr[i]);
			dico_add_string(d, DICO_OBJ_SECTION_WINATTR, (2*attrcnt)+0, winattr[i]);
			msgprintf(MSG_VERB2, "            dico_add_string(%s, winattr): key=%d, valsize=[%d]\n", relpath, (int)(2*attrcnt)+1, valsize);
			dico_add_data(d, DICO_OBJ_SECTION_WINATTR, (2*attrcnt)+1, value, valsize);
			attrcnt++;
		}
	}
	
	return 0;
}

int backup_item_stdattr(struct s_archive *ai, char *root, char *relpath, struct stat64 *statbuf, 
		struct s_dico *d, int *objtype, struct s_filesysdat *fsdat, u64 objectid)
{
	char fullpath[PATH_MAX];
	char buffer[PATH_MAX];
	int res;
	
	// init
	*objtype=OBJTYPE_NULL;
	concatenate_paths(fullpath, sizeof(fullpath), root, relpath);
	
	msgprintf(MSG_DEBUG2, "Adding [%.5ld]=[%s]\n", (long)objectid, relpath);
	if (dico_add_u64(d, DICO_OBJ_SECTION_STDATTR, DISKITEMKEY_OBJECTID, (u64)objectid)!=0)
	{	errprintf("dico_add_u64(DICO_OBJ_SECTION_STDATTR) failed\n");
		return -1;
	}
	if (dico_add_string(d, DICO_OBJ_SECTION_STDATTR, DISKITEMKEY_PATH, relpath)!=0)
	{	errprintf("dico_add_string(DICO_OBJ_SECTION_STDATTR) failed\n");
		return -1;
	}
	if (dico_add_u64(d, DICO_OBJ_SECTION_STDATTR, DISKITEMKEY_SIZE, (u64)statbuf->st_size)!=0)
	{	errprintf("dico_add_u64(DICO_OBJ_SECTION_STDATTR) failed\n");
		return -1;
	}
	if (dico_add_u32(d, DICO_OBJ_SECTION_STDATTR, DISKITEMKEY_MODE, (u32)statbuf->st_mode)!=0)
	{	errprintf("dico_add_u32(DICO_OBJ_SECTION_STDATTR) failed\n");
		return -1;
	}
	if (dico_add_u32(d, DICO_OBJ_SECTION_STDATTR, DISKITEMKEY_UID, (u32)statbuf->st_uid)!=0)
	{	errprintf("dico_add_u32(DICO_OBJ_SECTION_STDATTR) failed\n");
		return -1;
	}
	if (dico_add_u32(d, DICO_OBJ_SECTION_STDATTR, DISKITEMKEY_GID, (u32)statbuf->st_gid)!=0)
	{	errprintf("dico_add_u32(gid) failed\n");
		return -1;
	}
	if (dico_add_u64(d, DICO_OBJ_SECTION_STDATTR, DISKITEMKEY_ATIME, (u32)statbuf->st_atime)!=0)
	{	errprintf("dico_add_u32(DICO_OBJ_SECTION_STDATTR) failed\n");
		return -1;
	}
	if (dico_add_u64(d, DICO_OBJ_SECTION_STDATTR, DISKITEMKEY_MTIME, (u32)statbuf->st_mtime)!=0)
	{	errprintf("dico_add_u32(DICO_OBJ_SECTION_STDATTR) failed\n");
		return -1;
	}
	
	// 2. copy specific properties to the dico
	switch (statbuf->st_mode & S_IFMT)
	{
		case S_IFDIR:
			*objtype=OBJTYPE_DIR;
			break;
		case S_IFLNK:
			*objtype=OBJTYPE_SYMLINK;
			memset(buffer, 0, sizeof(buffer));
			if ((readlink(fullpath, buffer, sizeof(buffer)))<0)
				return -1;
			dico_add_string(d, DICO_OBJ_SECTION_STDATTR, DISKITEMKEY_SYMLINK, buffer);
			break;
		case S_IFREG:
			if (statbuf->st_nlink>1) // there are several links to that inode: there are hard links
			{
				res=dichl_get(fsdat->dichardlinks, (u64)statbuf->st_rdev, (u64)statbuf->st_ino, buffer, sizeof(buffer));
				if (res==0) // inode already seen --> hard link
				{	dico_add_string(d, DICO_OBJ_SECTION_STDATTR, DISKITEMKEY_HARDLINK, buffer);
					*objtype=OBJTYPE_HARDLINK;
				}
				else // next link to thar inode will be an hard link
				{
					dichl_add(fsdat->dichardlinks, (u64)statbuf->st_rdev, (u64)statbuf->st_ino, relpath);
				}
			}
			if (*objtype==OBJTYPE_NULL) // not an hard-link: it's a regular file or the first link when multiple links found
			{
				// don't allow reg-files with hardlinks to be copied as a small-file with other small-files
				// the file would be copied later in the archive (when the block for small files is ready)
				// and then the hardlink may come before the regfile (first link to that object)
				// case 1: file smaller than threshold: pack its data with other small files in a single compressed block
				if ((statbuf->st_size < g_options.smallfilethresh) && (statbuf->st_nlink==1))
					*objtype=OBJTYPE_REGFILEMULTI;
				else // case 2: file having hardlinks or larger than the threshold
					*objtype=OBJTYPE_REGFILEUNIQUE;
			}
			break;
		case S_IFCHR:
			dico_add_u64(d, DICO_OBJ_SECTION_STDATTR, DISKITEMKEY_RDEV, statbuf->st_rdev);
			*objtype=OBJTYPE_CHARDEV;
			break;
		case S_IFBLK:
			dico_add_u64(d, DICO_OBJ_SECTION_STDATTR, DISKITEMKEY_RDEV, statbuf->st_rdev);
			*objtype=OBJTYPE_BLOCKDEV;
			break;
		case S_IFIFO:
			dico_add_u64(d, DICO_OBJ_SECTION_STDATTR, DISKITEMKEY_RDEV, statbuf->st_rdev);
			*objtype=OBJTYPE_FIFO;
			break;
		case S_IFSOCK:
			dico_add_u64(d, DICO_OBJ_SECTION_STDATTR, DISKITEMKEY_RDEV, statbuf->st_rdev);
			*objtype=OBJTYPE_SOCKET;
			break;
		default:
			errprintf("unknown item %s\n", fullpath);
			return -1;
			break;
	}
	
	dico_add_u32(d, DICO_OBJ_SECTION_STDATTR, DISKITEMKEY_OBJTYPE, *objtype);
	
	return 0;
}

int backup_file(struct s_archive *ai, char *root, char *relpath, struct stat64 *statbuf, struct s_filesysdat *fsdat, u64 objectid)
{
	char fullpath[PATH_MAX];
	struct s_dico *dicoattr;
	int objtype;
	int res;

	// init	
	concatenate_paths(fullpath, sizeof(fullpath), root, relpath);
	
	// ---- backup file attributes (standard + xattr)
	if ((dicoattr=dico_alloc())==NULL)
	{	errprintf("dico_alloc() failed\n");
		return -1;
	}
	
	if (backup_item_stdattr(ai, root, relpath, statbuf, dicoattr, &objtype, fsdat, objectid)!=0)
	{	errprintf("backup_item_stdattr() failed: cannot prepare header for item %s\n", relpath);
		dico_destroy(dicoattr);
		return -1;
	}
	
	if (backup_item_xattr(ai, root, relpath, statbuf, dicoattr)!=0)
	{	errprintf("backup_item_xattr() failed: cannot prepare xattr-dico for item %s\n", relpath);
		dico_destroy(dicoattr);
		return -1;
	}
	
	if (filesys[fsdat->fstype].winattr==true)
	{
		if (backup_item_winattr(ai, root, relpath, statbuf, dicoattr)!=0)
		{	errprintf("backup_item_winattr() failed: cannot prepare winattr-dico for item %s\n", relpath);
			dico_destroy(dicoattr);
			return -1;
		}
	}
	
	/*if (objtype!=OBJTYPE_REGFILEMULTI) // multi-regfiles write all the headers and the block when block is ready
		queue_add_header(&g_queue, d, FSA_MAGIC_OBJT, fsdat->fsid);*/
	
	if (get_interrupted()==false)
		msgprintf(MSG_VERB1, "-[%.2d][%s] %s\n", fsdat->fsid, get_objtype_name(objtype), relpath);
	
	// ---- backup file contents for regfiles
	switch (objtype)
	{
		case OBJTYPE_REGFILEUNIQUE:
			if ((res=backup_obj_regfile_unique(ai, dicoattr, relpath, fullpath, statbuf->st_size, fsdat->fsid))!=0)
			{	errprintf("backup_obj_regfile_unique(%s)=%d failed\n", relpath, res);
				return -1;
			}
			break;
		case OBJTYPE_REGFILEMULTI:
			if ((res=backup_obj_regfile_multi(ai, dicoattr, relpath, fullpath, statbuf->st_size, fsdat->fsid, &fsdat->regmulti))!=0)
			{	errprintf("backup_obj_regfile_multi(%s)=%d failed\n", relpath, res);
				return -1;
			}
			break;
		default: // not a regular file
			queue_add_header(&g_queue, dicoattr, FSA_MAGIC_OBJT, fsdat->fsid);
			break;
	}
	
	return 0;
}

int backup_dir(struct s_archive *ai, char *root, char *path, struct s_filesysdat *fsdat, u64 dev, u64 *objectid)
{
	char fulldirpath[PATH_MAX];
	char fullpath[PATH_MAX];
	char relpath[PATH_MAX];
	struct stat64 statbuf;
	struct dirent *dir;
	DIR *dirdesc;
	int res;
	
	// init 
	concatenate_paths(fulldirpath, sizeof(fulldirpath), root, path);
	
	dirdesc = opendir(fulldirpath);
	if (!dirdesc)
	{       sysprintf("cannot open directory %s\n", fulldirpath);
		return -1;
	}
	
	// backup the directory itself (important for the root of the filesystem)
	if (lstat64(fulldirpath, &statbuf)!=0)
	{	sysprintf("cannot lstat64(%s)\n", fulldirpath);
		goto backup_dir_err;
	}
	if (backup_file(ai, root, path, &statbuf, fsdat, (*objectid)++)!=0)
	{	errprintf("backup_file(%s,%s) failed\n", root, path);
		goto backup_dir_err;
	}
	
	while (((dir = readdir(dirdesc)) != NULL) && (get_interrupted()==false))
	{
		// ---- calculate paths
		concatenate_paths(relpath, sizeof(relpath), path, dir->d_name);
		concatenate_paths(fullpath, sizeof(fullpath), fulldirpath, dir->d_name);
		if (lstat64(fullpath, &statbuf)!=0)
		{	sysprintf("cannot stat64(%s)\n", fullpath);
			closedir(dirdesc);
			return -1;
		}
		
		// ---- ignore "." and ".." and ignore mount-points
		if (strcmp(dir->d_name,".")==0 || strcmp(dir->d_name,"..")==0)
			continue; // ignore "." and ".."
		
		// ---- if dev is specified (when we backup a filesystem not a dir), ignore all other devices
		if ((dev!=0) && (u64)(statbuf.st_dev)!=dev)
			continue;
		
		// backup contents before the directory itself so that the dir-attributes are written after the dir contents
		if (S_ISDIR(statbuf.st_mode))
		{
			// if "dev==0" then accept to go on other filesystems (archtype=ARCHTYPE_DIRECTORIES)
			// we keep the same dichardlinks even if the filesystem is different (with different inodes numbers)
			// since the struct s_dichl is able to manage several filesystems (its key1)
			if ((dev==0) && (u64)(statbuf.st_dev)!=dev) // cross filesystems when archtype=ARCHTYPE_DIRECTORIES
			{
				res=backup_dir(ai, root, relpath, fsdat, statbuf.st_dev, objectid);
			}
			else // we are still on the same filesystem
			{
				res=backup_dir(ai, root, relpath, fsdat, dev, objectid);
			}
			if (res!=0)
			{	errprintf("backup_dir(%s) failed\n", relpath);
				goto backup_dir_err;
			}
		}
		else // not a directory
		{	
			if (backup_file(ai, root, relpath, &statbuf, fsdat, (*objectid)++)!=0)
			{	errprintf("backup_file(%s) failed\n", relpath);
				goto backup_dir_err;
			}
		}
	}
	
backup_dir_err:
	closedir(dirdesc);
	return 0;
}

int write_mainhead(struct s_archive *ai, char *label)
{
	u8 bufcheckclear[FSA_CHECKPASSBUF_SIZE+8];
	u8 bufcheckcrypt[FSA_CHECKPASSBUF_SIZE+8];
	struct md5_ctx md5ctx;
	struct timeval now;
	struct s_dico *d;
	u8 md5sum[16];
	u64 cryptsize;
	
	if (!ai)
	{	errprintf("ai is NULL\n");
		return -1;
	}
	
	// init
	gettimeofday(&now, NULL);
	ai->creattime=now.tv_sec;
	if ((d=dico_alloc())==NULL)
	{	errprintf("dico_alloc() failed\n");
		return -1;
	}
	
	dico_add_string(d, 0, MAINHEADKEY_FILEFORMATVER, FSA_FILEFORMAT);
	dico_add_string(d, 0, MAINHEADKEY_PROGVERCREAT, FSA_VERSION);
	dico_add_string(d, 0, MAINHEADKEY_ARCHLABEL, label);
	dico_add_u64(d, 0, MAINHEADKEY_CREATTIME, ai->creattime);
	dico_add_u32(d, 0, MAINHEADKEY_ARCHIVEID, ai->archid);
	dico_add_u32(d, 0, MAINHEADKEY_ARCHTYPE, ai->archtype);
	dico_add_u32(d, 0, MAINHEADKEY_COMPRESSALGO, ai->compalgo);
	dico_add_u32(d, 0, MAINHEADKEY_COMPRESSLEVEL, ai->complevel);
	dico_add_u32(d, 0, MAINHEADKEY_ENCRYPTALGO, ai->cryptalgo);
	dico_add_u32(d, 0, MAINHEADKEY_FSACOMPLEVEL, ai->fsacomp);
	
	if (ai->archtype==ARCHTYPE_FILESYSTEMS)
	{	
		dico_add_u64(d, 0, MAINHEADKEY_FSCOUNT, ai->fscount);
	}
	
	// if encryption is enabled, save the md5sum of a random buffer to check the password
#ifdef OPTION_CRYPTO_SUPPORT
	if (ai->cryptalgo!=ENCRYPT_NONE)
	{
		memset(md5sum, 0, sizeof(md5sum));
		generate_random_buffer(bufcheckclear, FSA_CHECKPASSBUF_SIZE);
		crypt_block_blowfish(FSA_CHECKPASSBUF_SIZE, &cryptsize, bufcheckclear, bufcheckcrypt, g_options.encryptpass, strlen((char*)g_options.encryptpass), true);
		
		md5_init_ctx(&md5ctx);
		md5_process_bytes(bufcheckclear, FSA_CHECKPASSBUF_SIZE, &md5ctx);
		md5_finish_ctx(&md5ctx, md5sum);
		
		assert(dico_add_data(d, 0, MAINHEADKEY_BUFCHECKPASSCLEARMD5, md5sum, 16)==0);
		assert(dico_add_data(d, 0, MAINHEADKEY_BUFCHECKPASSCRYPTBUF, bufcheckcrypt, FSA_CHECKPASSBUF_SIZE)==0);
	}
#endif // OPTION_CRYPTO_SUPPORT
	
	if (queue_add_header(&g_queue, d, FSA_MAGIC_MAIN, FSA_FILESYSID_NULL)!=0)
	{	errprintf("cannot write dico for main header\n");
		dico_destroy(d);
		return -1;
	}
	
	return 0;
}

int filesystem_save(struct s_archive *ai, char *partition, char *partmount, u16 fsid, int fstype, bool mountedbyfsa)
{
	struct s_dico *dicobegin=NULL;
	struct s_dico *dicoend=NULL;
	struct stat64 statbuf;
	struct s_filesysdat fsdat;
	u64 objectid=0;
	int ret=0;
	
	// write "begin of filesystem" header
	if ((dicobegin=dico_alloc())==NULL)
	{	errprintf("dicostart=dico_alloc() failed\n");
		return -1;
	}
	// TODO: add stats about files count in that dico
	queue_add_header(&g_queue, dicobegin, FSA_MAGIC_FSYB, fsid);
	
	// get dev ids of the filesystem to ignore mount points
	if (lstat64(partmount, &statbuf)!=0)
	{	sysprintf("cannot lstat64(%s)\n", partmount);
		return -1;
	}
	
	// init filesystemdata struct
	fsdat.fsid=fsid;
	fsdat.fstype=fstype;
	regmulti_init(&fsdat.regmulti, g_options.datablocksize);
	if ((fsdat.dichardlinks=dichl_alloc())==NULL)
	{	errprintf("dichardlinks=dichl_alloc() failed\n");
		return -1;
	}
	
	ret=backup_dir(ai, partmount, "/", &fsdat, (u64)statbuf.st_dev, &objectid);
	
	// put all small files that are in the last block to the queue
	if (regmulti_save_enqueue(&fsdat.regmulti, &g_queue, fsid)!=0)
	{	errprintf("Cannot queue last block of small-files\n");
		return -1;
	}
	
	// dico for hard links not required anymore
	dichl_destroy(fsdat.dichardlinks);
	
	// write "end of filesystem" header
	if ((dicoend=dico_alloc())==NULL)
	{	errprintf("dicoend=dico_alloc() failed\n");
		return -1;
	}
	// TODO: add stats about files count in that dico
	queue_add_header(&g_queue, dicoend, FSA_MAGIC_DATF, fsid);
	
	// unmount the partition if necessary
	if (mountedbyfsa==true)
	{
		msgprintf(MSG_VERB2, "unmounting [%s] which is mounted on [%s]\n", partition, partmount);
		if (filesys[fstype].umount(partition, partmount)!=0)
		{	sysprintf("cannot umount %s\n", partmount);
		}
		else
		{	rmdir(partmount); // remove temp dir created by fsarchiver
		}
	}
	
	return ret;
}

int filesystem_mount_partition(struct s_dico *dicofs, char *partition, char *partmnt, int *fstype, u16 fsid, bool *mntbyfsa)
{
	char fsbuf[FSA_MAX_FSNAMELEN];
	struct statvfs64 statfsbuf;
	struct s_strlist reqmntopt;
	struct s_strlist badmntopt;
	struct s_strlist curmntopt;
	int showwarningcount1=0;
	int showwarningcount2=0;
	char temp[PATH_MAX];
	char optbuf[128];
	u64 fsbytestotal;
	u64 fsbytesused;
	int readwrite;
	int count;
	int res;
	int i;
	
	res=generic_get_mntinfo(partition, &readwrite, temp, sizeof(temp), optbuf, sizeof(optbuf), fsbuf, sizeof(fsbuf));
	if (res==0) // partition is already mounted
	{
		*mntbyfsa=false;
		snprintf(partmnt, PATH_MAX, "%s", temp); // return the mount point to main savefs function
		msgprintf(MSG_DEBUG1, "generic_get_mntinfo(%s): mnt=[%s], opt=[%s], fs=[%s], rw=[%d]\n", partition, partmnt, optbuf, fsbuf, readwrite);
		if (readwrite==1 && g_options.allowsaverw==0)
		{
			errprintf("partition [%s] is mounted read/write. Please mount it read-only and \n"
				"try again. you can do \"mount -o remount,ro %s\"\n", partition, partition);
			return -1;
		}
		if (generic_get_fstype(fsbuf, fstype)!=0)
		{	errprintf("filesystem of partition [%s] is not supported by fsarchiver: [%s]\n", partition, fsbuf);
			return -1;
		}
		// check the filesystem is mounted with the right mount-options (to preserve acl and xattr)
		strlist_init(&reqmntopt);
		strlist_init(&badmntopt);
		strlist_init(&curmntopt);
		if (filesys[*fstype].reqmntopt(partition, &reqmntopt, &badmntopt)!=0)
		{
			errprintf("cannot get the required mount options for partition=[%s]\n", partition);
			strlist_empty(&reqmntopt);
			strlist_empty(&badmntopt);
			return -1;
		}
		strlist_split(&curmntopt, optbuf, ',');
		msgprintf(MSG_DEBUG2, "mount options found for partition=[%s]: [%s]\n", partition, strlist_merge(&curmntopt, temp, sizeof(temp), ','));
		msgprintf(MSG_DEBUG2, "mount options required for partition=[%s]: [%s]\n", partition, strlist_merge(&reqmntopt, temp, sizeof(temp), ','));
		msgprintf(MSG_DEBUG2, "mount options to avoid for partition=[%s]: [%s]\n", partition, strlist_merge(&badmntopt, temp, sizeof(temp), ','));
		count=strlist_count(&reqmntopt);
		for (i=0; i < count; i++)
		{
			strlist_getitem(&reqmntopt, i, optbuf, sizeof(optbuf));
			msgprintf(MSG_DEBUG2, "checking there is reqmntopt[%d]=[%s]\n", i, optbuf);
			if (strlist_exists(&curmntopt, optbuf)!=true)
			{
				if (showwarningcount1==0)
					errprintf("partition [%s] has to be mounted with options [%s] in order to preserve all its attributes\n",
						partition, strlist_merge(&reqmntopt, temp, sizeof(temp), ','));
				if (g_options.dontcheckmountopts==true)
				{	if (showwarningcount1++ == 0) // show this warning only once
						errprintf("fsarchiver will continue anyway since the option '-a' was used\n");
				}
				else // don't ignore the mount options
				{	errprintf("fsarchiver cannot continue, you can use option '-a' to ignore "
						"the mount options (xattr or acl may not be preserverd)\n");
					return -1;
				}
			}
		}
		count=strlist_count(&badmntopt);
		for (i=0; i < count; i++)
		{
			strlist_getitem(&badmntopt, i, optbuf, sizeof(optbuf));
			msgprintf(MSG_DEBUG2, "checking there is not badmntopt[%d]=[%s]\n", i, optbuf);
			if (strlist_exists(&curmntopt, optbuf)==true)
			{
				if (showwarningcount2==0)
					errprintf("partition [%s] has to be mounted without options [%s] in order to preserve all its attributes\n",
						partition, strlist_merge(&badmntopt, temp, sizeof(temp), ','));
				if (g_options.dontcheckmountopts==true)
				{	if (showwarningcount2++ == 0) // show this warning only once
						errprintf("fsarchiver will continue anyway since the option '-a' was used\n");
				}
				else // don't ignore the mount options
				{	errprintf("fsarchiver cannot continue, you can use option '-a' to ignore "
						"the mount options (xattr or acl may not be preserverd)\n");
					return -1;
				}
			}
		}
		strlist_empty(&reqmntopt);
		strlist_empty(&badmntopt);
		strlist_empty(&curmntopt);
	}
	else // partition not yet mounted
	{
		*mntbyfsa=true;
		mkdir_recursive(partmnt);
		msgprintf(MSG_DEBUG1, "partition %s is not mounted\n", partition);
		for (*fstype=-1, i=0; (filesys[i].name) && (*fstype==-1); i++)
		{
			if ((filesys[i].test(partition)==true) && (filesys[i].mount(partition, partmnt, filesys[i].name, MS_RDONLY)==0))
			{	*fstype=i;
				msgprintf(MSG_DEBUG1, "partition %s successfully mounted on [%s] as [%s]\n", partition, partmnt, filesys[i].name);
			}
		}
		if (*fstype==-1)
		{	errprintf("can't detect and mount filesystem of partition [%s], cannot continue.\n", partition);
			return -1;
		}
	}
	
	// get space statistics
	if (statvfs64(partmnt, &statfsbuf)!=0)
	{	errprintf("statvfs64(%s) failed\n", partmnt);
		return -1;
	}
	fsbytestotal=(u64)statfsbuf.f_frsize*(u64)statfsbuf.f_blocks;
	fsbytesused=fsbytestotal-((u64)statfsbuf.f_frsize*(u64)statfsbuf.f_bfree);
	
	dico_add_string(dicofs, 0, FSYSHEADKEY_FILESYSTEM, filesys[*fstype].name);
	dico_add_string(dicofs, 0, FSYSHEADKEY_MNTPATH, partmnt);
	dico_add_u64(dicofs, 0, FSYSHEADKEY_BYTESTOTAL, fsbytestotal);
	dico_add_u64(dicofs, 0, FSYSHEADKEY_BYTESUSED, fsbytesused);
	
	if (filesys[*fstype].getinfo(dicofs, partition)!=0)
	{	errprintf("cannot save filesystem attributes for partition %s\n", partition);
		return -1;
	}
	
	return 0;
}

int do_create(char *archive, char **partition, int archtype)
{
	struct s_dico *dicofsinfo[FSA_MAX_FSPERARCH];
	char partmounts[PATH_MAX][FSA_MAX_FSPERARCH];
	pthread_t thread_comp[FSA_MAX_COMPJOBS];
	bool mountedbyfsa[FSA_MAX_FSPERARCH];
	int fstype[FSA_MAX_FSPERARCH];
	char fullpath[PATH_MAX];
	char currentdir[PATH_MAX];
	pthread_t thread_writer;
	struct s_dico *dicoend;
	struct s_archive ai;
	struct stat64 st;
	u64 objectid=0;
	char *curdir;
	int itemcount;
	int ret=0;
	int i;
	
	// init archive
	archive_init(&ai);
	archive_generate_id(&ai);

	// init misc data struct to zero
	for (i=0; i<FSA_MAX_COMPJOBS; i++)
		thread_comp[i]=0;
	for (i=0; i<FSA_MAX_FSPERARCH; i++)
		fstype[i]=-1;
	for (i=0; i<FSA_MAX_FSPERARCH; i++)
		memset(partmounts[i], 0, PATH_MAX);
	thread_writer=0;
	
	// check that partitions are all block devices
	for (i=0; partition[i]; i++)
	{
		if (archtype==ARCHTYPE_FILESYSTEMS)
		{
			if ((stat64(partition[i], &st)!=0) || (!S_ISBLK(st.st_mode)))
			{	errprintf("%s is not a valid block device\n", partition[i]);
				ret=-1;
				goto do_create_error;
			}
		}
	}
	
	// pass options to archive
	path_force_extension(ai.basepath, PATH_MAX, archive, ".fsa");
	ai.cryptalgo=g_options.encryptalgo;
	ai.compalgo=g_options.compressalgo;
	ai.complevel=g_options.compresslevel;
	ai.fsacomp=g_options.fsacomplevel;
	ai.fscount=itemcount=i;
	ai.archtype=archtype;
	
	// create compression threads
	for (i=0; (i<g_options.compressjobs) && (i<FSA_MAX_COMPJOBS); i++)
	{
		if (pthread_create(&thread_comp[i], NULL, thread_comp_fct, (void*)&ai) != 0)
		{       errprintf("pthread_create(thread_comp_fct) failed\n");
			ret=-1;
			goto do_create_error;
		}
	}
	
	// create archive-writer thread
	if (pthread_create(&thread_writer, NULL, thread_writer_fct, (void*)&ai) != 0)
	{       errprintf("pthread_create(thread_writer_fct) failed\n");
		ret=-1;
		goto do_create_error;
	}
	
	// write archive main header
	if (write_mainhead(&ai, "archive-label")!=0)
	{       errprintf("archive_write_mainhead(%s) failed\n", archive);
		ret=-1;
		goto do_create_error;
	}
	
	// write one fsinfo header for each filesystem
	if (archtype==ARCHTYPE_FILESYSTEMS)
	{
		for (i=0; (i < itemcount) && (partition[i]); i++)
		{
			if ((dicofsinfo[i]=dico_alloc())==NULL)
			{	errprintf("dico_alloc() failed\n");
				return -1;
			}
			generate_random_tmpdir(partmounts[i], PATH_MAX, i);
			if (filesystem_mount_partition(dicofsinfo[i], partition[i], partmounts[i], &fstype[i], i, &mountedbyfsa[i])!=0)
			{	msgprintf(MSG_STACK, "archive_filesystem(%s) failed\n", partition[i]);
				goto do_create_error;
			}
			if (queue_add_header(&g_queue, dicofsinfo[i], FSA_MAGIC_FSIN, FSA_FILESYSID_NULL)!=0)
			{	errprintf("queue_add_header(FSA_MAGIC_FSIN, %s) failed\n", partition[i]);
				goto do_create_error;
			}
		}
	}
	
	// copy contents to archive
	if (archtype==ARCHTYPE_FILESYSTEMS) // write contents of each filesystem
	{
		for (i=0; (i < itemcount) && (partition[i]!=NULL) && (get_interrupted()==false); i++)
		{
			msgprintf(MSG_VERB1, "============= archiving filesystem %s =============\n", partition[i]);
			if (filesystem_save(&ai, partition[i], partmounts[i], i, fstype[i], mountedbyfsa[i])!=0)
			{	errprintf("archive_filesystem(%s) failed\n", partition[i]);
				goto do_create_error;
			}
		}
	}
	else if (archtype==ARCHTYPE_DIRECTORIES)
	{
		struct s_filesysdat fsdat;
		
		// init filesystemdata struct
		fsdat.fsid=0;
		fsdat.fstype=0; // TODO: fstype when multiple fs saved in one dir
		regmulti_init(&fsdat.regmulti, g_options.datablocksize);
		
		if ((fsdat.dichardlinks=dichl_alloc())==NULL)
		{	errprintf("dichardlinks=dichl_alloc() failed\n");
			goto do_create_error;
		}
		
		for (i=0; (i < itemcount) && (partition[i]) && (get_interrupted()==false); i++)
		{
			curdir=partition[i];
			if (curdir[0]=='/') // absolute path
			{
				snprintf(fullpath, sizeof(fullpath), "%s", curdir);
				backup_dir(&ai, "/", fullpath, &fsdat, (u64)0, &objectid); 
			}
			else // relative path
			{
				concatenate_paths(fullpath, sizeof(fullpath), getcwd(currentdir, sizeof(currentdir)), curdir);
				backup_dir(&ai, ".", curdir, &fsdat, (u64)0, &objectid);
				// put all small files that are in the last block to the queue
			}
		}
		
		// put all small files that are in the last block to the queue
		if (regmulti_save_enqueue(&fsdat.regmulti, &g_queue, 0)!=0)
		{	errprintf("Cannot queue last block of small-files\n");
			return -1;
		}
		
		// dico for hard links not required anymore
		dichl_destroy(fsdat.dichardlinks);
		
		// write "end of file" header
		if ((dicoend=dico_alloc())==NULL)
		{	errprintf("dicoend=dico_alloc() failed\n");
			goto do_create_error;
		}
		// TODO: add stats about files count in that dico
		queue_add_header(&g_queue, dicoend, FSA_MAGIC_DATF, FSA_FILESYSID_NULL);
	}
	else // error
	{	errprintf("unsupported archtype: %d\n", archtype);
		goto do_create_error;
	}
	
	if (get_interrupted()==false)
		goto do_create_success;
	
do_create_error:
	msgprintf(MSG_DEBUG1, "THREAD-MAIN1: exit error\n");
	ret=-1;
	
	for (i=0; (i < FSA_MAX_FSPERARCH) && (partmounts[i]); i++)
	{
		if (mountedbyfsa[i]==true)
		{
			generic_umount(partmounts[i]);
			rmdir(partmounts[i]);
		}
	}

do_create_success:
	msgprintf(MSG_DEBUG1, "THREAD-MAIN1: exit\n");
	queue_set_end_of_queue(&g_queue, true); // other threads must not wait for more data from this thread
	
	for (i=0; (i<g_options.compressjobs) && (i<FSA_MAX_COMPJOBS); i++)
		if (thread_comp[i] && pthread_join(thread_comp[i], NULL) != 0)
			errprintf("pthread_join(thread_comp) failed\n");
	
	if (thread_writer && pthread_join(thread_writer, NULL) != 0)
		errprintf("pthread_join(thread_writer) failed\n");
	
	if (ret!=0)
		archive_remove(&ai);
	
	archive_destroy(&ai);
	return ret;
}
